// File objQueue.h for CC1B
#ifndef _OBJQUEUE_H_
#define _OBJQUEUE_H_
//include other libraries here
#include <system/memory.h>
#pragma library 1

/**
 * \file objQueue.h
 * \brief Library for queue object
 *
 * This library provides a queue object that can store up to 14 bytes. The queue must reside in a single\n
 * rambank. The functions are written so that they can be used in any call tree (interrupt, mainlevel\n
 * and tasklevel) by using respectively _IsrTemp, _MainTemp or _TaskTemp as the ram location.\n
 * Any character array that resides in a single rambank can be initialized to be a queue.
 *
 * \code
 * char q[12];
 * _MainTemp = 8; \\size of the queue must be stored in a global ram location
 * objQueue_init(q,_MainTemp);
 * \endcode
 *
 * Bytes are stored into the queue by using function objQueue_enqueue().\n
 * This function requires the value to be enqueued to be stored in a global ram location.
 * \code
 * if (!objQueue_isFull(q)) {
 *   _MainTemp = value; //value to enqueu must be in a global ram location
 *   objQueue_enqueue(q,_MainTemp);
 * }
 * \endcode
 *
 * Bytes are retrieved from the queue by using function objQueue_dequeue().\n
 * The retrieved byte is returned in W if the queue is not empty.
 * \code
 * if (!objQueue_isEmpty(q)) {
 *   value = objQueue_dequeue(q);
 * }
 * \endcode
 */

//queue buffer layout
//
// +----+----+
// |size| num|  size = number of databytes in queue buffer, num = number of values in queue 
// +----+----+
// |head|tail|  head = position last stored value, tail = position before value to retrieve
// +----+----+
// | d a t a |
// | d a t a |

/**
 * \brief Initialize a queue
 *
 * The function objQueue_init() initializes a character array as a queue object. The character array\n
 * must reside in a single rambank. The overhead for a queue object is 2 bytes. So for a queue that can\n
 * hold 6 bytes, a size of 8 bytes must be specified.
 *
 * \n
 * How to use:
 * \code
 *   char q[12];
 *   _MainTemp = 12; //size of queue must be in a global ram location
 *   objQueue_init(q,_MainTemp);
 * \endcode
 *
 * \param queue Queue address
 * \param size Size of the queue. The size MUST have been stored in a global ram location (0x00-0x0F).\n
 * Valid range for size is 3 to 16 (depends on queue address). The queue cannot cross a rambank boundary.
 * \n
 */
#define objQueue_init(queue,size) \
{ \
  W = queue; \
#pragma update_FSR 0 \
  FSR = W; \
  size--; \
  size--; \
  W = swap(size); \
  INDF = W; \
  FSR++; \
  INDF = 0; \
#pragma update_FSR 1 \
}

/**
 * \brief Check if queue is empty
 *
 * The function objQueue_isEmpty() checks wether a queue holds any data.\n
 * Use this function together with objQueue_dequeue() to be sure a queue holds any data.
 *
 * \n
 * How to use:
 * \code
 * if (!objQueue_isEmpty(q)) {
 *   value = objQueue_dequeue(q);
 * }
 * \endcode
 *
 * \param W Queue address.
 * \return True (1) if queue empty, false (0) if not empty.
 * \n
 */
bit objQueue_isEmpty(char W)
{
#pragma update_FSR 0
  FSR = W;
  W = 0x0F;
  W &= INDF; //extract num
#pragma update_FSR 1
  return Zero_;
}

/**
 * \brief Check if queue is full
 *
 * The function objQueue_isFull() checks wether a queue can hold another byte.\n
 * Use this function together with objQueue_enqueue() to be sure a queue hold more data.
 *
 * \n
 * How to use:
 * \code
 * if (!objQueue_isFull(q)) {
 *   _MainTemp = value; //value to enqueue must be in a global ram location
 *   objQueue_enqueue(q,_MainTemp);
 * }
 * \endcode
 *
 * \param W Queue address.
 * \return True (1) if queue full, false (0) if not full.
 * \n
 */
bit objQueue_isFull(char W)
{
#pragma update_FSR 0
  FSR = W;
  W = swap(INDF);
  W ^= INDF; //xor num with size
#pragma update_FSR 1
  return Zero_;
}

/**
 * \brief Clear queue
 *
 * The function 0bjQueue_clear() clears a queue, in effect removing any data it may hold.\n
 * This is established by resetting its internal pointers.
 *
 * \n
 * How to use:
 * \code
 * objQueue_clear(q);
 * \endcode
 *
 * \param W Queue address.
 * \n
 */
void objQueue_clear(char W)
{
#pragma update_FSR 0
  FSR = W;
  W = 0xF0;
  INDF &= W;  //clear num
  FSR++;
  INDF = 0; //clear headtail
#pragma update_FSR 1
}

/**
 * \brief Get queue length
 *
 * The function objQueue_length() calculates the number of bytes stored in the queue.\n
 * This can be up to the capacity of the queue.
 *
 * \n
 * How to use:
 * \code
 * length = objQueue_length(q);
 * \endcode
 *
 * \param W Queue address.
 * \return Number of stored bytes (0 to queu capacity).
 * \n
 */
char objQueue_length(char W)
{
#pragma update_FSR 0
  FSR = W;
  W = INDF; //get num in lownib
  W &= 0x0F;
#pragma update_FSR 1
  return W;
}

/**
 * \brief Get queue capacity
 *
 * The function objQueue_capacity() calculates the maximum number of bytes that can be\n
 * stored into the queue. This capacity equals the queue size minus the overhead (2 bytes).
 *
 * \n
 * How to use:
 * \code
 * capacity = objQueue_capacity(q);
 * \endcode
 *
 * \param W Queue address.
 * \return Capacity of queue (maximum 14).
 * \n
 */
char objQueue_capacity(char W)
{
#pragma update_FSR 0
  FSR = W;
  W = swap(INDF); //get size in lownib
  W &= 0x0F;
#pragma update_FSR 1
  return W;
}

/**
 * \brief Get queue size
 *
 * The function objQueue_size() calculates the size of a queue.\n
 * This is the size as specified during the last initialization of the queue.
 *
 * \n
 * How to use:
 * \code
 * size = objQueue_size(q);
 * \endcode
 *
 * \param W Queue address.
 * \return Size of queue (maximum 16).
 * \n
 */
char objQueue_size(char W)
{
#pragma update_FSR 0
  FSR = W;
  W = swap(INDF); //get size in lownib
  W &= 0x0F;
  FSR = 0;
  FSR.1 = 1;  //FSR=2 (overhead)
  W += FSR; //get total queue size
#pragma update_FSR 1
  return W;
}

/**
 * \brief Get queue free space
 *
 * The function objQueue_free() calculates how many additional bytes can be stored in\n
 * a queue before it is full. This free space equals capacity minus length.
 *
 * \n
 * How to use:
 * \code
 * free = objQueue_free(q);
 * \endcode
 *
 * \param W Queue address.
 * \return Number of additional bytes the queue can hold (0 to queu capacity).
 * \n
 */
char objQueue_free(char W)
{
#pragma update_FSR 0              //example:
  FSR = W;                        // INDF = 0101_0010 = size_num  size-num = 5-2=3
  W = swap(INDF); //get size in lownib         // W       = 0010_0101
  W = INDF - W;                   // W       = 0010_1101
  W &= 0x0F;                      // W       = 0000_1101  W=13
  FSR = 0;                        // FSR    = 0
  FSR.4 = 1;                      // FSR    = 16
  W = FSR - W;                    // W       = 16-13 = 3
#pragma update_FSR 1
  return W;
}

/**
 * \brief Get enqueue address
 *
 * INTERNAL FUNCTION.\n
 * The function _objQueue_enqueueAddress() checks if a queue can hold another byte and\n
 * adjusts pointers if so. This function should only be called from the macro objQueue_enqueue().\n
 * This function adjusts the queue's head pointer and num field if the queue is not full.
 *
 * \n
 * How to use:
 * \code
 * FSR = _objQueue_enqueueAddress(q);
 * INDF = value; //value must a global ram location
 * \endcode
 *
 * \param W Queue address.
 * \return Enqueue address (if queue is not full).
 * \n
 */
static char _objQueue_enqueueAddress(char W)
{
#pragma update_FSR 0
  FSR = W;
  W = swap(INDF);
  W ^= INDF; // xor size with num
  if (!Zero_) {
    INDF++; //inc num
    FSR++;
    INDF += 0x10; //inc head
    FSR--;
    W = INDF; //get size
    FSR++;
    W ^= INDF; //xor size with head
    W &= 0xF0;
    W = 0x0F;
    if (Zero_) INDF &= W; //if head == size then clear head
    W = swap(INDF); //get head in lownib
    W &= 0x0F;
    FSR++; //point to queue data begin
    W += FSR; //head position
  }
#pragma update_FSR 1
  return W;
}

/**
 * \brief Enqueue byte into queue
 *
 * The function objQueue_enqueue() stores a byte into a queue. If the queue is already\n
 * full, the byte is not stored. This is to protect the queue integrity. To know that a byte\n
 * is stored, this function should be used together with the objQueue_isFull() function.
 *
 * \n
 * How to use:
 * \code
 * if (!objQueue_isFull(q)) {
 *   _MainTemp = value; //value to enqueue must be in a global ram location
 *   objQueue_enqueue(q,_MainTemp);
 * }
 * \endcode
 *
 * \param queue Queue address.
 * \param value Byte to enqueue. The value MUST have been stored in a global ram location (0x00-0x0F).
 * \n
 */
#define objQueue_enqueue(queue,value) \
{ \
  FSR = _objQueue_enqueueAddress(queue); \
  INDF = value; \
}

/**
 * \brief Dequeue byte from queue
 *
 * The function objQueue_dequeue() retrieves a byte from a queue. If the queue is already\n
 * empty, a void value is returned. This is to protect the queue integrity. To know that a byte\n
 * is retrieved, this function should be used together with the objQueue_isEmpty() function.
 *
 * \n
 * How to use:
 * \code
 * if (!objQueue_isEmpty(q)) {
 *   value = objQueue_dequeue(q);
 * }
 * \endcode
 *
 * \param W Queue address.
 * \return Byte retrieved from queue (if queue is not empty).
 * \n
 */
char objQueue_dequeue(char W)
{
#pragma update_FSR 0
  FSR = W;
  W = INDF; //load size+num
  W &= 0x0F; //test num
  if (!Zero_) {
    INDF--; //dec num
	W = swap(INDF); //get size in lownib
	FSR++;
	INDF++; //inc tail
	W ^= INDF; //xor size with tail
	W &= 0x0F;
	W = 0xF0;
	if (Zero_) INDF &= W; //if tail == size then clear tail
	W = INDF; //get tail in lownib
	W &= 0x0F;
	FSR++; //point to queue data begin
	FSR += W; //tail position
	W = INDF; //get value
  }
#pragma update_FSR 1
  return W;
}

#pragma library 0
#endif
// End of file objQueue.h
