// File objStack.h for CC1B
#ifndef _OBJSTACK_H_
#define _OBJSTACK_H_
//include other libraries here
#include <system/memory.h>
#pragma library 1

/**
 * \file objStack.h
 * \brief Library for stack object
 *
 * This library provides a stack object that can store up to 15 bytes. The stack must reside in a single\n
 * rambank. The functions are written so that they can be used in any call tree (interrupt, mainlevel\n
 * and tasklevel) by using respectively _IsrTemp, _MainTemp or _TaskTemp as the ram location.\n
 * Any character array that resides in a single rambank can be initialized to be a stack.
 *
 * \code
 * char s[12];
 * _MainTemp = 8; \\size of the stack must be stored in a global ram location
 * objStack_init(s,_MainTemp);
 * \endcode
 *
 * Bytes are pushed onto the stack by using function objStack_push().\n
 * This function requires the value to be pushed to be stored in a global ram location.
 * \code
 * if (!objStack_isFull(s)) {
 *   _MainTemp = value; //value to push must be in a global ram location
 *   objStack_push(s,_MainTemp);
 * }
 * \endcode
 *
 * Bytes are popped from the stack by using function objStack_pop().\n
 * The popped byte is returned in W if the stack is not empty.
 * \code
 * if (!objStack_isEmpty(s)) {
 *   value = objStack_pop(s);
 * }
 * \endcode
 */

//stack buffer layout
//
// +----+----+
// |size| num|  size = number of databytes in stack buffer, num = number of values on stack 
// +----+----+
// | d a t a |
// | d a t a |

/**
 * \brief Initialize a stack
 *
 * The function objStack_init() initializes a character array as a stack object. The character array\n
 * must reside in a single rambank. The overhead for a stack object is 1 byte. So for a stack that can\n
 * hold 6 bytes, a size of 7 bytes must be specified.
 *
 * \n
 * How to use:
 * \code
 *   char s[12];
 *   _MainTemp = 12; //size of stack must be in a global ram location
 *   objStack_init(s,_MainTemp);
 * \endcode
 *
 * \param stack Stack address
 * \param size Size of the stack. The size MUST have been stored in a global ram location (0x00-0x0F).\n
 * Valid range for size is 2 to 16 (depends on stack address). The stack cannot cross a rambank boundary.
 * \n
 */
#define objStack_init(stack,size) \
{ \
  W = stack; \
#pragma update_FSR 0 \
  FSR = W; \
  size--; \
  W = swap(size); \
  INDF = W; \
#pragma update_FSR 1 \
}

/**
 * \brief Check if stack is empty
 *
 * The function objStack_isEmpty() checks wether a stack holds any data.\n
 * Use this function together with objStack_pop() to be sure a stack holds any data.
 *
 * \n
 * How to use:
 * \code
 * if (!objStack_isEmpty(s)) {
 *   value = objStack_pop(s);
 * }
 * \endcode
 *
 * \param W Stack address.
 * \return True (1) if stack empty, false (0) if not empty.
 * \n
 */
bit objStack_isEmpty(char W)
{
#pragma update_FSR 0
  FSR = W;
  W = 0x0F;
  W &= INDF; //extract num
#pragma update_FSR 1
  return Zero_;
}

/**
 * \brief Check if stack is full
 *
 * The function objStack_isFull() checks wether a stack can hold another byte.\n
 * Use this function together with objStack_push() to be sure a stack hold more data.
 *
 * \n
 * How to use:
 * \code
 * if (!objStack_isFull(s)) {
 *   _MainTemp = value; //value to push must be in a global ram location
 *   objStack_push(s,_MainTemp);
 * }
 * \endcode
 *
 * \param W Stack address.
 * \return True (1) if stack full, false (0) if not full.
 * \n
 */
bit objStack_isFull(char W)
{
#pragma update_FSR 0
  FSR = W;
  W = swap(INDF);
  W ^= INDF; //xor num with size
#pragma update_FSR 1
  return Zero_;
}

/**
 * \brief Clear stack
 *
 * The function 0bjStack_clear() clears a stack, in effect removing any data it may hold.\n
 * This is established by resetting its internal pointers.
 *
 * \n
 * How to use:
 * \code
 * objStack_clear(s);
 * \endcode
 *
 * \param W Stack address.
 * \n
 */
void objStack_clear(char W)
{
#pragma update_FSR 0
  FSR = W;
  W = 0xF0;
  INDF &= W;  //clear num
#pragma update_FSR 1
}

/**
 * \brief Get stack length
 *
 * The function objStack_length() calculates the number of bytes stored in the stack.\n
 * This can be up to the capacity of the stack.
 *
 * \n
 * How to use:
 * \code
 * length = objStack_length(s);
 * \endcode
 *
 * \param W Stack address.
 * \return Number of stored bytes (0 to stack capacity).
 * \n
 */
char objStack_length(char W)
{
#pragma update_FSR 0
  FSR = W;
  W = INDF; //get num in lownib
  W &= 0x0F;
#pragma update_FSR 1
  return W;
}

/**
 * \brief Get stack capacity
 *
 * The function objStack_capacity() calculates the maximum number of bytes that can be\n
 * pushed onto the stack. This capacity equals the stack size minus the overhead (1 byte).
 *
 * \n
 * How to use:
 * \code
 * capacity = objStack_capacity(s);
 * \endcode
 *
 * \param W Stack address.
 * \return Capacity of stack (maximum 15).
 * \n
 */
char objStack_capacity(char W)
{
#pragma update_FSR 0
  FSR = W;
  W = swap(INDF); //get size in lownib
  W &= 0x0F;
#pragma update_FSR 1
  return W;
}

/**
 * \brief Get stack size
 *
 * The function objStack_size() calculates the size of a stack.\n
 * This is the size as specified during the last initialization of the stack.
 *
 * \n
 * How to use:
 * \code
 * size = objStack_size(s);
 * \endcode
 *
 * \param W Stack address.
 * \return Size of stack (maximum 16).
 * \n
 */
char objStack_size(char W)
{
#pragma update_FSR 0
  FSR = W;
  W = swap(INDF); //get size in lownib
  W &= 0x0F;
  FSR = 0;
  FSR.0 = 1;  //FSR=1
  W += FSR; //get total stack size
#pragma update_FSR 1
  return W;
}

/**
 * \brief Get stack free space
 *
 * The function objStack_free() calculates how many additional bytes can be pushed\n
 * onto a stack before it is full. This free space equals capacity minus length.
 *
 * \n
 * How to use:
 * \code
 * free = objStack_free(s);
 * \endcode
 *
 * \param W Stack address.
 * \return Number of additional bytes the stack can hold (0 to stack capacity).
 * \n
 */
char objStack_free(char W)
{
#pragma update_FSR 0              //example:
  FSR = W;                        // INDF = 0101_0010 = size_num  size-num = 5-2=3
  W = swap(INDF); //get size in lownib         // W       = 0010_0101
  W = INDF - W;                   // W       = 0010_1101
  W &= 0x0F;                      // W       = 0000_1101  W=13
  FSR = 0;                        // FSR    = 0
  FSR.4 = 1;                      // FSR    = 16
  W = FSR - W;                    // W       = 16-13 = 3
#pragma update_FSR 1
  return W;
}

/**
 * \brief Get push address
 *
 * INTERNAL FUNCTION.\n
 * The function _objStack_pushAddress() checks if a stack can hold another byte and\n
 * adjusts pointers if so. This function should only be called from the macro objStack_push().\n
 * This function adjusts the stack's num field if the stack is not full.
 *
 * \n
 * How to use:
 * \code
 * FSR = _objStack_pushAddress(s);
 * INDF = value; //value must a global ram location
 * \endcode
 *
 * \param W Stack address.
 * \return Push address (if stack is not full).
 * \n
 */
static char _objStack_pushAddress(char W)
{
#pragma update_FSR 0
  FSR = W;
  W = swap(INDF); //load size+num
  W ^= INDF; //xor size with num
  if (!Zero_) {
    INDF++; //inc num
    W = INDF; //get num in lownib
	W &= 0x0F;
    FSR++; //point to stack data begin
    W += FSR; //top of stack position
  }
#pragma update_FSR 1
  return W;
}

/**
 * \brief Push byte onto stack
 *
 * The function objStack_push() pushes a byte onto a stack. If the stack is already\n
 * full, the byte is not stored. This is to protect the stack integrity. To know that a byte\n
 * is pushed, this function should be used together with the objStack_isFull() function.
 *
 * \n
 * How to use:
 * \code
 * if (!objStack_isFull(s)) {
 *   _MainTemp = value; //value to push must be in a global ram location
 *   objStack_push(s,_MainTemp);
 * }
 * \endcode
 *
 * \param stack Stack address.
 * \param value Byte to push. The value MUST have been stored in a global ram location (0x00-0x0F).
 * \n
 */
#define objStack_push(stack,value) \
{ \
  FSR = _objStack_pushAddress(stack); \
  INDF = value; \
}

/**
 * \brief Pop byte from stack
 *
 * The function objStack_pop() pops a byte from a stack. If the stack is already\n
 * empty, a void value is returned. This is to protect the stack integrity. To know that a byte\n
 * is popped, this function should be used together with the objStack_isEmpty() function.
 *
 * \n
 * How to use:
 * \code
 * if (!objStack_isEmpty(s)) {
 *   value = objStack_pop(s);
 * }
 * \endcode
 *
 * \param W Stack address.
 * \return Byte popped from stack (if stack is not empty).
 * \n
 */
char objStack_pop(char W)
{
#pragma update_FSR 0
  FSR = W;
  W = INDF; //load size+num
  W &= 0x0F; //test num
  if (!Zero_) {
    INDF--; //dec num
    W = INDF; //get num in lownib
    W &= 0x0F;
    FSR++; //point to stack data begin
    FSR += W; //top of stack position
    W = INDF; //get value
  }
#pragma update_FSR 1
  return W;
}

#pragma library 0
#endif
// End of file objStack.h
