// File Task.h for CC1B
#ifndef _TASK_H_
#define _TASK_H_
//include other libraries here
#pragma library 1

/**
 * \file Task.h
 * \brief Library for taskswitching support
 *
 * The taskswitch mechanism is NOT multitasking. It allows subroutines to run automatically at specified\n
 * intervals. It can best be compared to the Quick Basic ON TIMER GOSUB command. The interrupt\n
 * routine monitors when it is time to run a task. Then a contextswitch is done, so that after the\n
 * interrupt routine exits, a task routine is executed, rather than the mainline code that was interrupted.\n
 * The interrupts remain active while the task runs. Only after the task exits, will the mainline code\n
 * continue (after another contextswitch). It is therefore imperative that a task never blocks or waits\n
 * for an event. If it must wait, let it set a flag and return. The next time it runs, it can check the flag.\n
 * The SX secret instructions are used to perform the contextswitch.
 *
 * \n
 * For clarity, TASK is defined as void. It allows you to use
 * \code
 * TASK myTask()
 * {
 * }
 * \endcode
 *
 * Note that task routines appear to CC1B as part of the interrupt call tree, although the task routines\n
 * themselves execute as mainlevel code. You should never call a TASK subroutine from mainline code.
 *
 * \n
 * Here are some guide 'rules' to determine the correct tasks parameters.\n
 * The goal is to LET EACH TASK START ON TIME.
 *
 * \n
 * For each task to run on time, each task must return within 1 tasktick and the minimum\n
 * interval must be equal to the number of tasks. To guarantee a minimum percentage of \n
 * CPU cycles for the mainline code, you must use an interval greater than the number of tasks.
 * \code
 * #define  TASKS  4
 * #define  TASKINTERVAL  5
 * \endcode
 * This guarantees ((5-4)/5) * 100% = 20% of CPU cycles is used for mainline code.
 *
 * \n
 * For each task to run on time, the individual task intervals must be a multiple of TASKINTERVAL.\n
 * If you specify for some task an interval of 2*TASKINTERVAL, then each 2nd timeslice for\n
 * that task is not used to run the task but used for mainline code. If a task does not return within\n
 * 1 tasktick, either split up the task in smaller pieces (state machine) or increase the TASKTICK value.\n
 * So the rule is:
 *
 * \n
 * <b>All tasks must have an interval equal to K*TASKINTERVAL and must return within 1 tasktick\n
 * and should start 1 tasktick apart. This ensures tasks do not overlap and start on time.</b>
 *
 * \n
 * You are not bound by the above rule. You can set the interval to any non-zero value (1-255) but then\n
 * it may happen that tasks are at some point scheduled for the same time and then the tasks are not\n
 * deterministic. Tasks scheduled for the same time are simply executed after each other without\n
 * returning to the mainline code inbetween.
 *
 * \n
 * If K>1 then a task does not use all its timeslices. Unused timeslices are not available to other\n
 * tasks but are used for mainline code. If a task takes too much time, a statemachine can be\n
 * used to make a task return in time. If a task must run at a lower rate than set by its interval\n
 * parameter, the task can decrement a counter and return immediately if it has not reached 0.
 *
 * \n
 * Since the interval parameter is an 8bit value, the highest K calculates from K*TASKINTERVAL = 255.\n
 * This yields Kmax = int(255/TASKINTERVAL)\n
 * INTPERIOD determines the interrupt tick: isrtick = INTPERIOD/CPUFREQ\n
 * TASKTICK determines the task tick: tasktick = TASKTICK*isrtick\n
 * The maximum value for TASKTICK is 65279.\n
 * TASKINTERVAL must be >= TASKS. If you don't define TASKINTERVAL\n
 * then TASKINTERVAL will  be set to TASKS.
 *
 * \n
 * Example: suppose you want to run a task 1000 times per second, with given\n
 * CPUFREQ=20_000_000 and INTPERIOD=217 and 4 tasks running:\n
 * isrtick = INTPERIOD/CPUFREQ = 217/20_000_000 = 10.85 uSec\n
 * tasktick = TASKTICK * 10.85 uSec = (1/1000) Sec / 4 = 250 uSec\n
 * TASKTICK = 250/10.85 = 23 for interval = 4.\n
 * Normally the task that must run at the highest rate determines TASKTICK\n
 * and task intervals are adjusted for tasks that run at a lower rate.\n
 * \n
 */

#ifndef TASKS
#error TASKS (1-5) needs to be defined before including <Task.h>
#else
#if (TASKS < 1) || (TASKS > 5)
#error TASKS needs to be in range 1 to 5
#endif
#endif

#ifndef TASKINTERVAL
#define TASKINTERVAL (TASKS)
#else
#if TASKINTERVAL < TASKS
#error TASKINTERVAL must be >= TASKS
#endif
#endif

#ifndef TASKLIST
#error TASKLIST needs to be defined before including <Task.h>
#else
#if (16 - (TASKLIST & 0x0F)) < (3*TASKS+1)
#error TASKLIST value too high, TaskList array not in single bank
#endif
#endif
 
#define TASK void
#define TASKSTART 0x80
#define TASKRUNONCE 0x40

static char TaskList[3*TASKS+1] @TASKLIST;
  static char TaskId[TASKS] @TaskList[0];
  static char TaskInterval[TASKS] @TaskList[TASKS];
  static char TaskRunTick[TASKS] @TaskList[2*TASKS];
  static char TaskToRun @TaskList[3*TASKS];

#define TASKKERNEL 0xF0
static char TaskCountLSB @0xF0; //16bits divisor to keep track when contextswitch must occur
static char TaskCountMSB @0xF1;
static char TaskDelay @0xF2; //number of taskticks to wait before task runs, or number of taskticks passed while task runs
static char TaskSuspendTicks @0xF3; //number of taskticks that a task can suspend itself
static char Task_M @0xF4; //flags and save area for M
  static bit TaskRunning @Task_M.5; //set if task code is currently running
  static bit TaskFinished @Task_M.6; //set if task code finished
  static bit TaskEnabled @Task_M.7; //set if tasks are allowed to run
static char TaskGlobalTemp @0xFA; //mainlevel code save area
static char Task_W @0xFB; 
static char Task_ST @0xFC;
static char Task_FSR @0xFD;
static char Task_PCL @0xFE; // mainlevel code return address
static char Task_PCH @0xFF;

#pragma cdata.TASKDATA
#if TASKS == 1
#pragma cdata[] = 0
#pragma cdata[] = TASKINTERVAL
#pragma cdata[] = 1
#elif TASKS == 2
#pragma cdata[] = 0,0
#pragma cdata[] = TASKINTERVAL,TASKINTERVAL
#pragma cdata[] = 1,2
#elif TASKS == 3
#pragma cdata[] = 0,0,0
#pragma cdata[] = TASKINTERVAL,TASKINTERVAL,TASKINTERVAL
#pragma cdata[] = 1,2,3
#elif TASKS == 4
#pragma cdata[] = 0,0,0,0
#pragma cdata[] = TASKINTERVAL,TASKINTERVAL,TASKINTERVAL,TASKINTERVAL
#pragma cdata[] = 1,2,3,4
#elif TASKS == 5
#pragma cdata[] = 0,0,0,0,0
#pragma cdata[] = TASKINTERVAL,TASKINTERVAL,TASKINTERVAL,TASKINTERVAL,TASKINTERVAL
#pragma cdata[] = 1,2,3,4,5
#endif
#pragma cdata[] = 0

/**
 * \brief Initialize the task scheduler
 *
 * The function TaskInit() copies the task initialization data from rom to ram.\n
 * It is to be called once from mainlevel code only and it must be called before\n
 * any of the other task functions.
 *
 * \n
 * How to use:
 * \code
 * TaskInit();
 * \endcode
 * \n
 */
void TaskInit(void)
{
  RomCopy(TaskList,TASKDATA,3*TASKS+1);
  TaskCountLSB = 1;
  TaskCountMSB = 1;
}

/**
 * \brief Disable the task scheduler
 *
 * The function TaskDisable() prevents all tasks from running when called from mainline code.\n
 * It pauses the task timer. If however it is called from a task subroutine then this subroutine\n
 * does not need to return within 1 tasktick. This may be handy in the case of a high priority event.\n
 * When used in a task subroutine, the task should call TaskEnable() before returning to allow\n
 * other tasks to run, otherwise no other task will run after returning to mainline code.
 *
 * \n
 * How to use:
 * \code
 * TASK myTask()
 * {
 *   if (highPriority) {
 *     TaskDisable();
 *     HandleEvent(); //this may take longer than 1 tasktick
 *     TaskEnable();
 *   }
 * }
 * \endcode
 * \n
 */
void TaskDisable(void)
{
  TaskEnabled = 0;
}

/**
 * \brief Enable the task scheduler
 *
 * The function TaskEnable() allows tasks to run. It starts the task timer. It does not reset the\n
 * task timer, it simply continues to count, so to tasks it appears time was frozen while TaskDisable()\n
 * was active. After calling this function, all tasks that are scheduled to run, will be run.
 *
 * \n
 * \code
 * TaskEnable();
 * \endcode
 * \n
 */
void TaskEnable(void)
{
  TaskEnabled = 1;
}

/**
 * \brief Install task in tasklist
 *
 * The function TaskSet() sets up the task parameters in the TaskList.\n
 * The parameter taskid may be ORed with TASKSTART (to autostart task)\n
 * and TASKRUNONCE (so the task only runs once).
 *
 * \n
 * \param slot Index in TaskList (0 to TASKS-1).
 * \param taskid Identifier for task subroutine (may be ORed with TASKSTART and/or TASKRUNONCE).
 * \param interval Specifies in taskticks (>=1) how frequently the task runs. Do not use value 0.
 * \n
 */
#define TaskSet(slot,taskid,interval) \
{ \
  TaskInterval[slot] = interval; \
  TaskId[slot] = taskid; \
}

/**
 * \brief Locate slot for task subroutine
 *
 * The function TaskSlot() locates in which slot a specific task is installed.\n
 * If the task is found, its slot id is returned, otherwise 0xFF.
 *
 * \n
 * \param W Task identifier.
 * \return Slot in which task is installed, or 0xFF if task not found (eg. task not set).
 * \n
 */
char TaskSlot(char W)
{
  _MainTemp = W;
  FSR = TaskId;
#pragma update_FSR 0
next:
  W = TaskId;
  W = FSR - W;
  W ^= TASKS;
  if (Zero_) goto done;
  W = INDF;
  W ^= _MainTemp;
  W &= 0x3F;
  if (Zero_) { //slot for task found
    W = TaskId - 1;
	W = FSR - W;
	goto done;
  }
  FSR++;
  goto next;
done:
#pragma update_FSR 1
  _MainTemp = W;
  return _MainTemp + 255;
}

/**
 * \brief Locate available task slot
 *
 * The function TaskSlotAvailable() locates a free task slot.\n
 * If available, its id is returned, otherwise 0xFF.
 *
 * \n
 * \return Slot id for available slot, or 0xFF (no slot available).
 * \n
 */
char TaskSlotAvailable(void)
{
  FSR = TaskId;
#pragma update_FSR 0
next:
  W = TaskId;
  W = FSR - W;
  W ^= TASKS;
  if (Zero_) goto done;
  W = INDF;
  W &= 0x3F;
  if (Zero_) { //free slot found
    W = TaskId - 1;
	W = FSR - W;
	goto done;
  }
  FSR++;
  goto next;
done:
#pragma update_FSR 1
  _MainTemp = W;
  return _MainTemp + 255;
}

/**
 * \brief Get task timer value
 *
 * The function TaskTimer() returns the current task timer count. This is an 8bit value.\n
 * The counter simply increments every tasktick, except when TaskDisable() is called.\n
 * Then the timer is paused until TaskEnable() is called.
 *
 * \n
 * \return Current task timer value.
 * \n
 */
char TaskTimer(void)
{
#pragma update_FSR 0
  FSR = TaskRunTick;
  W = TaskToRun;
  FSR += W;
  W = INDF;         // W = TaskRunTick[TaskToRun]
  FSR = 0;
  FSR = ~FSR;       // FSR =0xFF (eg. bank TASKKERNEL)
  if (TaskRunning) {
    // TaskDelay holds passed ticks for current task -> runtime = TaskRunTick[TaskToRun] + TaskDelay
    W += TaskDelay; // W = TaskRunTick[TaskTorun] + TaskDelay
  }
  else {
    // TaskDelay holds ticks that must pass before next task runs -> runtime= TaskRunTick[TaskToRun] - TaskDelay
	W = TaskDelay-W; // W = TaskDelay - TaskRunTick[TaskToRun]
    W = ~W;          // W = ~(TaskDelay - TaskRunTick[TaskToRun])
    FSR = 0;
    FSR++;           // FSR = 1
    W += FSR;        // W = ~(TaskDelay - TaskRunTick[TaskToRun]) + 1 = TaskRunTick[TaskToRun] - TaskDelay
  }
#pragma update_FSR 1
  return W;
}

/**
 * \brief Start a task
 *
 * The function TaskStart() enables a task that has been installed in the tasklist.\n
 * If the task is not in the tasklist nothing happens, otherwise the task will run\n
 * at the time it is scheduled for.
 *
 * \n
 * \param W Task identifier.
 * \n
 */
void TaskStart(char W)
{
  W = TaskSlot(W);
  FSR = W;
  if (!FSR.7) {
    FSR += TaskId;
    INDF.7 = 1;
  }
}

/**
 * \brief Stop a task
 *
 * The function TaskStop() disables a task that has been installed in the tasklist.\n
 * If the task is not in the tasklist nothing happens, otherwise the task will not run\n
 * until it is enabled again by TaskStart() or TaskSlotStart(), but it will continue\n
 * to reschedule according to its interval parameter, so not to upset the task sequence.
 *
 * \n
 * \param W Task identifier.
 * \n
 */
void TaskStop(char W)
{
  W = TaskSlot(W);
  FSR = W;
  if (!FSR.7) {
    FSR += TaskId;
    INDF.7 = 0;
  }
}

/**
 * \brief Start a task
 *
 * The function TaskSlotStart() enables the task that is installed in the specified slot.\n
 * If no task is specified for that slot nothing happens, otherwise the task will run\n
 * at the time it is scheduled for.
 *
 * \n
 * \param W Slot id (0 to TASKS-1).
 * \n
 */
void TaskSlotStart(char W)
{
  FSR = W;
  FSR += TaskId;
  INDF.7 = 1;
}

/**
 * \brief Stop a task
 *
 * The function TaskSlotStop() disables the task that is installed in the specified slot.\n
 * The task will not run until it is enabled again by TaskStart() or TaskSlotStart(), but\n
 * it will continue to reschedule according to its interval parameter, so not to upset\n
 * the task sequence.
 *
 * \n
 * \param W Slot id (0 to TASKS-1).
 * \n
 */
void TaskSlotStop(char W)
{
  FSR = W;
  FSR += TaskId;
  INDF.7 = 0;
}

/**
 * \brief Reschedule a task
 *
 * The function TaskReschedule() reschedules the task this function is called from.\n
 * When called from mainline code nothing happens. It is used to allow a task to change\n
 * the rate it is run, due to some condition, for example the arrival of data. The task\n
 * will be rescheduled using the parameter rather than its interval value. The interval\n
 * value however is not changed.
 * 
 * \n
 * How to use:
 * \code
 * TASK mytask()
 * {
 *   //some code
 *   TaskReschedule(12); //run this task again, 12 taskticks after this task returns
 * }
 * \endcode
 *
 * \param W Number of taskticks after which the task runs again once it returns.\n
 * A value of 0 will use the task interval value for rescheduling.
 * \n
 */
void TaskReschedule(char W)
{
  TaskSuspendTicks = W;
}

/**
 * \brief Copy bytes from rom to ram
 *
 * The function TaskRomCopy() copies bytes from rom to ram. These bytes are the lower 8 bits\n
 * of the rom words. No checks are done on validity of rom and ram addresses.\n
 * Call this function only from TASK subroutines. For a simular function for mainline code\n
 * see RomCopy() in the memory.h library.
 *
 * \n
 * How to use:
 * \code
 * TaskRomCopy(dest,source,len);
 * \endcode
 *
 * \param dest Physical ram address.
 * \param source Physical rom address (SX18/20/28: 0x000-0x7FF, SX48/52: 0x000-0xFFF).
 * \param len Number of bytes to copy.
 * \n
 */
void TaskRomCopy(char dest, long source, char len)
{
  char i = 0;
  while (len > 0) {
    _MainTemp = i; //set index for LargeArrayAddress()
	char k = LargeArrayAddress(dest);
    RomWord(source.high8,source.low8);
    _MainTemp = W;
	_SystemRam[k] = _MainTemp;
	source++;
	i++;
	len--;
  }
}

#pragma library 0

/**
 * \brief Taskswitch handler (extern)
 *
 * The function Taskswitch() must be supplied by the application.\n
 * A typical layout for this function:
 * \code
 * #define TXDEQUEUE 1 //unique identifiers for task subroutines
 * #define RXENQUEUE 2
 * void TaskSwitch(char W)
 * {
 *   switch (W & 0x3F) { //call task, returns within 1 tasktick
 *     case TXDEQUEUE: TxDequeue(); break; //TxDequeue has identifier 0
 *     case RXENQUEUE: RxEnqueue(); break; //RxEnqueue has identifier 1
 *   }
 * }
 * \endcode
 *
 * \param W Task identifier.
 * \n
 */
extern void TaskSwitch(char W); //application supplied function

/**
 * \brief The mainlevel part of the task scheduler
 *
 * The function TaskSchedule() is called by the interrupt part of the task scheduler.\n
 * It should never be called from anywhere else. It executes as mainlevel code.\n
 * This function calls the function TaskSwitch that must be supplied by the application.\n
 * It is important to realize that this function in effect interrupts the mainline code.
 *
 * \n
 * The application must supply a function TaskSwitch() that must look like this:
 * \code
 * #define MYTASK1 12  //unique identifiers for tasks (1 to 63)
 * #define MYTASK2 25
 *
 * TASK myTask1()
 * {
 * }
 *
 * TASK myTask2()
 * {
 * }
 *
 * void TaskSwitch(char W)
 * {
 *   switch (W & 0x3F) { //call task, returns within 1 tasktick
 *     case MYTASK1: myTask1(); break;
 *     case MYTASK2: myTask2(); break;
 *     //more tasks here
 *   }
 * }
 * \endcode
 *
 * Note that the functions TaskSwitch() and the task subroutines all\n
 * execute as mainlevel code. They do not consume interrupt cycles.
 * \n
 */
void TaskSchedule(void)
{
  char index,runtime,delay;
  // start task code (interrupts mainlevel code, executes outside isr)
  nop(); //in case interrupt occured in middle of skip instruction
  FSR = TASKKERNEL;
#pragma update_FSR 0
  Task_M &= 0xE0;
  W = MODE; //save m
  W &= 0x1F;
  Task_M |= W;
  W = _MainTemp; //save _MainTemp
  TaskGlobalTemp = W;
#pragma update_FSR 1
Task_run:
  TaskSuspendTicks = 0; //may be set by task using TaskReschedule()
  index = TaskId[TaskToRun];
  if (index.7 == 1) { //only run task if enabled
    TaskSwitch(index);
	if (index.6) TaskId[TaskToRun].7 = 0; //if runonce, stop task after run
  }
  //reschedule current task
  runtime = TaskRunTick[TaskToRun]; //current time
  W = TaskSuspendTicks;
  if (!W) {
	W = TaskInterval[TaskToRun];
  }
  _MainTemp = W + runtime;
  TaskRunTick[TaskToRun] = _MainTemp;
  //find next task to run
  delay = 255;
  for (index=0; index<TASKS; index++) {
	W = TaskRunTick[index];
    _MainTemp = W - runtime;
    if (_MainTemp < delay) {
      delay = _MainTemp;
      TaskToRun = index; //next time that TaskSchedule runs, the task in TaskToRun is called
	}
  }
  if (delay <= TaskDelay) { //next task was scheduled for same time, or for later time but taskticks are missed so it must run now
    TaskDelay = TaskDelay - delay;
    goto Task_run; //no contextswitch, but immediately run the next task
  }
  else {
    TaskDelay = delay - TaskDelay; //next task scheduled for later time
  }
  // exit task code (contextswitch to run mainlevel code)
#pragma update_FSR 0
  FSR = TASKKERNEL;
  W = TaskGlobalTemp;
  _MainTemp = W;
  W = Task_M; //restore m
  MODE = W;
  W = 25; //check if rtcc will overrun when doing contextswitch directly
  //-------------------------
  W += RTCC; // 21+4 cycles are required to do context switch now (leaves new rtcc @mainlevel <= $FC)
  TaskFinished = 1;
  if (Carry) { //carry clear if enough cycles left to do context switch now
Task_exit_wait:
    goto Task_exit_wait; //wait for interrupt to do contextswitch
  }
  //task has finished, restore registers from outside isr
  W = Task_W; //restore w
#asm
  DW 0x48 ;PUSHW ;1
#endasm
  W = Task_ST; //restore st
#asm
  DW 0x49 ;PUSHST ;1
#endasm
  W = Task_FSR; //restore fsr
#asm
  DW 0x4A ;PUSHFSR ;1
#endasm
  W = Task_PCH; //restore pc
  MODE = W;
  W = Task_PCL;
#asm
  DW 0x4B ;PUSHPC ;1
#endasm
  W = Task_M; //restore m
  MODE = W;
  TaskRunning = 0;
  TaskFinished = 0;
#asm
  DW 0x0E ;RETI ;3 //continue to execute mainlevel code (leaving rtcc <= $FC)
#endasm
#pragma update_FSR 1
}

/**
 * \brief The interrupt part of the task scheduler
 *
 * The function Task_ISR() must run at the end of the interrupt routine. It must not be called\n
 * anywhere else. Once installed, if a task must run, it redirects program control to TaskSchedule().
 *
 * \n
 * How to use:
 * \code
 * interrupt iServer()
 * {
 * #if defined _CHIP_SX18_ || defined _CHIP_SX20_ || defined _CHIP_SX28_
 * _IsrMode = MODE; //save M
 * #endif
 * //----------
 * //other interrupt code here
 * //----------
 * Task_ISR; //interrupt part of the task scheduler
 * #if defined _CHIP_SX18_ || defined _CHIP_SX20_ || defined _CHIP_SX28_
 * MODE = _IsrMode; //restore M
 * #endif
 * W = -INTPERIOD;
 * retiw();
 * }
 * \endcode
 * \n
 */
#define Task_ISR \
{ \
  FSR = TASKKERNEL; \
  if (0) TaskSchedule(); \
#pragma update_FSR 0 \
  if (!TaskEnabled) goto Task_contextswitch; \
  TaskCountLSB--; \
  if (!Zero_) goto Task_contextswitch; \
#if TASKTICK > 255 \
  TaskCountMSB--; \
  if (!Zero_) goto Task_contextswitch; \
  TaskCountMSB = (TASKTICK+256) >> 8; \
#endif \
  TaskCountLSB = (TASKTICK+256) & 0xFF; \
  TaskDelay++; \
  if (TaskRunning) goto Task_contextswitch; \
  TaskDelay--; \
  if (!Zero_) { \
    TaskDelay--; \
  } \
  if (TaskDelay) goto Task_contextswitch; \
  TaskRunning = 1; \
#asm \
  DW 0x4C ;POPW \
#endasm \
  Task_W = W; \
#asm \
  DW 0x4D ;POPST \
#endasm \
  Task_ST = W; \
#asm \
  DW 0x4E ;POPFSR \
#endasm \
  Task_FSR = W; \
#asm \
  DW 0x4F ;POPPC \
#endasm \
  Task_PCL = W; \
  W = MODE; \
  Task_PCH = W; \
  W = 0; \
  MODE = W; \
  W = 0x02; \
#asm \
  DW 0x4B ;PUSHPC ;retiw will force contextswitch to run TaskSchedule() \
#endasm \
  goto Task_exit; \
Task_contextswitch: \
  if (!TaskFinished) goto Task_exit; \
  W = Task_W; \
#asm \
  DW 0x48 ;PUSHW \
#endasm \
  W = Task_ST; \
#asm \
  DW 0x49 ;PUSHST \
#endasm \
  W = Task_FSR; \
#asm \
  DW 0x4A ;PUSHFSR \
#endasm \
  W = Task_PCH; \
  MODE = W; \
  W = Task_PCL; \
#asm \
  DW 0x4B ;PUSHPC ;retiw will force contextswitch to run mainlevel code \
#endasm \
  TaskRunning = 0; \
  TaskFinished = 0; \
Task_exit: \
#pragma update_FSR 1 \
}

#endif
// End of file Task.h
