// File vplib.h for CC1B

/*
	This library requires the following constants to be defined before including the library
	
	#define  VPSLOTS  4     //set number of VP slots (SX18/20/28: 1 to 6, SX48/52: 1 to 8)
	#define  VPRAM  0x50  //set first vp ram bank
*/

#ifndef _VPLIB_H_
#define _VPLIB_H_
//include other libraries here
#pragma library 1

/**
 * \file vplib.h
 * \brief Library for virtual peripheral support
 *
 * This library provides a generic interface to virtual peripherals. Basic elements of a\n
 * virtual peripheral are a set of data and functions that access that data. Some VP's\n
 * have interrupt code, or mainline code or both. One property of this generic interface\n
 * is that the VP dataset can be located in any of the rambanks assigned to VP's.\n
 * This makes it possible to use dynamic loading of VP's. This means VP's can be\n
 * installed and uninstalled as needed.
 * \n
 */

#ifndef VPSLOTS
#error VPSLOTS must be defined before including <vplib.h>
#endif
#ifndef VPRAM
#error VPRAM must be defined before including <vplib.h>
#endif

/*
 * vpListType array holds  entries of vp type identifying data
 * vpListDefs array holds offset to label VPDEFINITIONS
 * vpList index number specifies rambank: 0 ->first vp object,VPSLOTS-1 -> last vp object
 * a vpListType entry value of 0 specifies a free entry
*/
#if defined _CHIP_SX18_ || defined _CHIP_SX20_ || defined _CHIP_SX28_
#define VPLIST (VPRAM + (VPSLOTS<<5))
#endif
#if defined _CHIP_SX48_ || defined _CHIP_SX52_
#define VPLIST (VPRAM + (VPSLOTS<<4))
#endif

//vpList arrays
static shadowDef char vpListType[VPSLOTS] @ VPLIST;
static shadowDef char vpListDefs[VPSLOTS] @ VPLIST + VPSLOTS;

#if defined _CHIP_SX18_ || defined _CHIP_SX20_ || defined _CHIP_SX28_
#if VPSLOTS >= 0
#define VPBANK0 (VPRAM+0x00)
static char vpBank0[16] @ VPBANK0;
#endif
#if VPSLOTS >= 1
#define VPBANK1 (VPRAM+0x20)
static char vpBank1[16] @ VPBANK1;
#endif
#if VPSLOTS >= 2
#define VPBANK2 (VPRAM+0x40)
static char vpBank2[16] @ VPBANK2;
#endif
#if VPSLOTS >= 3
#define VPBANK3 (VPRAM+0x60)
static char vpBank3[16] @ VPBANK3;
#endif
#if VPSLOTS >= 4
#define VPBANK4 (VPRAM+0x80)
static char vpBank4[16] @ VPBANK4;
#endif
#if VPSLOTS >= 5
#define VPBANK5 (VPRAM+0xA0)
static char vpBank5[16] @ VPBANK5;
#endif
#endif

#if defined _CHIP_SX48_ || defined _CHIP_SX52_
#if VPSLOTS >= 0
#define VPBANK0 (VPRAM+0x00)
static char vpBank0[16] @ VPBANK0;
#endif
#if VPSLOTS >= 1
#define VPBANK1 (VPRAM+0x10)
static char vpBank1[16] @ VPBANK1;
#endif
#if VPSLOTS >= 2
#define VPBANK2 (VPRAM+0x20)
static char vpBank2[16] @ VPBANK2;
#endif
#if VPSLOTS >= 3
#define VPBANK3 (VPRAM+0x30)
static char vpBank3[16] @ VPBANK3;
#endif
#if VPSLOTS >= 4
#define VPBANK4 (VPRAM+0x40)
static char vpBank4[16] @ VPBANK4;
#endif
#if VPSLOTS >= 5
#define VPBANK5 (VPRAM+0x50)
static char vpBank5[16] @ VPBANK5;
#endif
#if VPSLOTS >= 6
#define VPBANK6 (VPRAM+0x60)
static char vpBank6[16] @ VPBANK6;
#endif
#if VPSLOTS >= 7
#define VPBANK7 (VPRAM+0x70)
static char vpBank7[16] @ VPBANK7;
#endif
#if VPSLOTS >= 8
#define VPBANK8 (VPRAM+0x80)
static char vpBank8[16] @ VPBANK8;
#endif
#endif

/**
 * \brief Initialize VP environment
 *
 * The function vpInit() initializes the vpList.\n
 * It must be called prior to any other function from this library.
 * \n
 */
void vpInit(void) {
  char i;
  for (i=0; i<VPSLOTS; i++) {
    vpListType[i] = 0;
  }
}

/**
 * \brief Setup pins for virtual peripheral
 *
 * The function _vpPinSetup() initializes pins for virtual peripherals.\n
 * Do not call this function directly from the application code.\n
 * The virtual peripheral pin setup data is stored in rom at the passed address.\n
 * This data consists of 3 byte records identifying pins and configuration.\n
 * If the first byte of this record is 0 then there are no more pins to setup.\n
 *
 * \param addr Address of virtual peripheral pin setup data
 * \param install 1 to setup pin after intall, 0 to setup pin after uninstall
 */
static void _vpPinSetup(long addr, char install) {
  char port, pinmask, pinconfig;
  while (1) {
    RomWord(addr.high8,addr.low8);
    port = W;
	if (!port) return;
	addr++;
	RomWord(addr.high8,addr.low8);
	pinmask = W;
	addr++;
	RomWord(addr.high8,addr.low8); //pinconfig
	pinconfig = W;
	if (install) pinconfig = swap(pinconfig);
	if (pinconfig.2) {
	  _MainTemp = pinmask;
	  if (pinconfig.1) {
	    setPinInput(port,_MainTemp);
	  }
	  if (pinconfig.0) {
	    setPinHigh(port,_MainTemp);
	  }
	  else {
	    setPinLow(port,_MainTemp);
	  }
	  if (!pinconfig.1) {
	    setPinOutput(port,_MainTemp);
	  }
	}
    addr++;
  }
}

/**
 * \brief Install virtual peripheral
 *
 * The function vpInstall() installs a VP into the first free entry of the vpList.\n
 * If all vp slots are occupied, this function returns 0.
 *
 * \n
 * How to use:
 * \code
 * char vph; //handle for vp
 * vph = vpInstall(MYVP);
 * //use the VP for some time
 * //until no longer needed
 * vpUninstall(vph);
 * \endcode
 *
 * \param addr Address of virtual peripheral initialization data.
 * \return handle for installed virtual peripheral if succesful, or 0 if not succesful.\n
 * The format of the handle:\n
 * - high nibble (b7-b4) are bits b7-b4 of rambank where virtual peripheral is installed\n
 * - low nibble (b3-b0) are the index in the vpListType array
 *
 * \n
 */
char vpInstall(long addr) {
  char i,len;
  for (i=0; i<VPSLOTS; i++) {
    if (vpListType[i] == 0) {
      _MainTemp = (char)(addr - VPDEFINITIONS);
	  vpListDefs[i] = _MainTemp;
	  RomWord(addr.high8,addr.low8);
	  _MainTemp = W;
	  vpListType[i] = _MainTemp;
      len = i << 4;
#if defined _CHIP_SX18_ || defined _CHIP_SX20_ || defined _CHIP_SX28_
      len <<= 1;
#endif
      len += VPRAM;
	  i += len; //rambank in high nibble, index in low nibble
      addr++;
      RomWord(addr.high8,addr.low8);
      len = W;
      addr++;
      RomCopy(i & 0xF0,addr,len);
	  addr += len; //point to pinconfig
      _vpPinSetup(addr,1);
	  return i;
	}
  }
  return 0;
}

/**
 * \brief Uninstall virtual peripheral
 *
 * The function vpUninstall() removes an installed VP from the vpList.\n
 * The handle must be considered invalid after using this function.\n
 * A new VP can be installed by calling vpInstall().
 *
 * \n
 * How to use:
 * \code
 * char vph; //handle for vp
 * vph = vpInstall(MYVP);
 * //use the VP for some time
 * //until no longer needed
 * vpUninstall(vph);
 * \endcode
 *
 * \param bnk Handle of installed VP.
 * \n
 */
void vpUninstall(char bnk) {
  bnk &= 0x0F;
  if (vpListType[bnk] != 0) {
    long addr = vpListDefs[bnk];
	addr += VPDEFINITIONS;
    vpListType[bnk] = 0;
    addr++;
    _vpPinSetup(addr,0);
  }
}

#pragma library 0
#endif
// End of file vplib.h
